/*
	The administrator-friendly (TM) LOCadmin client
*/

#include <stdio.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <termios.h>

#include "common.h"

// The infamous

char *StrDup (char *c)
{
	return strcpy (new char [strlen (c) + 1], c);
}

static unsigned int nslookup (char *h)
{
	struct hostent *ho;

	if (!(ho = gethostbyname (h)))
		return 0;

	return *(unsigned int*)*ho->h_addr_list;
}

static int connect_tcp (unsigned int IP, short port)
{
	struct sockaddr_in server;
	int Socket;

	if ((Socket = socket (AF_INET, SOCK_STREAM, 0)) == -1)
	{
		perror ("Can't create socket");
		exit (1);
	}

	server.sin_family = AF_INET;
	server.sin_port = htons (port);
	server.sin_addr.s_addr = IP;

	if (connect (Socket, (struct sockaddr*)&server,
	             sizeof (struct sockaddr)) == -1)
	{
		return -1;
	}

	return Socket;
}

#define SERVER_TIMEOUT 40*1000

static int getmessage (char *ret, int fd)
{
	struct pollfd pfd  = { fd, POLLIN, 0 };
	int i;

	i = poll (&pfd, 1, SERVER_TIMEOUT);
	if (i != 1) return 0;

	i = read (fd, ret, MAX_MSG-1);
	if (i == MAX_MSG-1) i = 0;

	return i;
}

static bool greet_LOC (int fd)
{
	int i;
	char *argv [MAX_ARGV], buffer [MAX_MSG];

	i = getmessage (buffer, fd);
	i = parseLOCmessage (buffer, i, argv);
	return i == 2 && MSGtype (argv [0]) == VERSION;
}

static char *admin_pass;

static int connect_LOC_admin (char *host, short port)
{
	unsigned int IP;
	int fd, i;
	md5string M;
	char *argv [MAX_ARGV], buffer [MAX_MSG];

	if ((IP = nslookup (host)) == 0) return -2;

	if ((fd = connect_tcp (IP, port)) == -1) return -3;
	if (!greet_LOC (fd)) return -4;

	i = mkLOCmessage (buffer, ADMIN, 0, NULL);
	write (fd, buffer, i);
	i = getmessage (buffer, fd);
	i = parseLOCmessage (buffer, i, argv);
	if (i != 2 && MSGtype (argv [0]) != AUTHSTR)
		return -4;

	argv [0] = do_md5digest (admin_pass, argv [1], M);
	i = mkLOCmessage (buffer, DIGEST, 1, argv);
	write (fd, buffer, i);
	i = getmessage (buffer, fd);
	i = parseLOCmessage (buffer, i, argv);

	return (i == 1 && MSGtype (argv [0]) == OKEY) ? fd : -1;
}

static int pass_prompt ()
{
	termios ini, hid;
	char inpass [20];

	tcgetattr (fileno (stdin), &ini);
	hid = ini;
	hid.c_lflag &= ~ECHO;
	tcsetattr (fileno (stdin), TCSAFLUSH, &hid);

	printf ("Enter Administrator Password: ");

	fgets (inpass, 19, stdin);
	tcsetattr (fileno (stdin), TCSANOW, &ini);

	inpass [strlen (inpass) - 1] = 0;
	if (inpass [0] == 0) return -1;

	admin_pass = StrDup (inpass);

	return 0;
}

const char instructions [] =
	"Reminder:-\n"
	"  a <username> <password>    : for creating a new user\n"
	"  d <username>               : for deleting a user\n"
	"  c <username> <password>    : for changing a user's password\n"
	"  ?                          : for this\n";
const char prompt [] = "Action: ";

#define MAX_CMD 200

int main (int argc, char **Argv)
{
	char *c, *s1, *s2 = NULL;
	int fd, i = 0;
	char command [MAX_CMD+1], *argv [MAX_ARGV], buffer [MAX_MSG];

	initLOCMessages ();

	if (argc == 1 || !(c = strchr (Argv [1], ':'))) {
		printf ("Usage: LOCadmin <server:port>\n");
		return 1;
	}

	pass_prompt ();
	*c = 0;

	if ((fd = connect_LOC_admin (Argv [1], atoi (c + 1))) < 0)
		switch (fd) {
		case -1:
			printf ("Authentication Failed! :(\n");
			return 1;
		case -2:
			printf ("Can't locate host [%s]\n", Argv [1]);
			return 1;
		case -3:
			printf ("Server not responding\n");
			return 1;
		case -4:
			printf ("Protocol failure. Not a LOC server\n");
			return 1;
		}

	printf ("Ok boss\n");

	puts (instructions);
	for (;;) {
		printf ("%s", prompt);
		if (!fgets (command, MAX_CMD, stdin)) break;
		command [strlen (command) - 1] = 0;
		if (command [0] == '?') {
			puts (instructions);
			continue;
		}
		if (command [0] != 'a' && command [0] != 'd'
		&& command [0] != 'c') {
			puts (instructions);
			continue;
		}
		if (!(s1 = strchr (command, ' '))) continue;
		while (*s1 == ' ') ++s1;
		if (command [0] != 'd') {
			if (!(s2 = strchr (s1, ' '))) continue;
			*s2++ = 0;
			while (*s2 == ' ') ++s2;
		}
		argv [0] = s1; argv [1] = s2;
		switch (command [0]) {
		case 'a':
			i = mkLOCmessage (buffer, AD_ADDUSER, 2, argv);
			break;
		case 'd':
			i = mkLOCmessage (buffer, AD_DELUSER, 1, argv);
			break;
		case 'c':
			i = mkLOCmessage (buffer, AD_CHPASS, 2, argv);
			break;
		}
		write (fd, buffer, i);
		i = getmessage (buffer, fd);
		i = parseLOCmessage (buffer, i, argv);
		if (i == 0) {
			printf ("Connection to server closed\n");
			break;
		}
		switch (MSGtype (argv [0])) {
		case OKEY:
			printf ("Ok Mastah\n");
			break;
		case USEREXISTS:
			printf ("That user already exists boss\n");
			break;
		case NOSUCHUSER:
			printf ("That user does not have an account boss\n");
			break;
		case BADPARAMETER:
			printf ("A string too long or invalid your highness\n");
			break;
		default:
			printf ("Unexpected reply [%s]\n", argv [0]);
			goto wofl;
		}
	}
wofl:

	close (fd);
}
