#include "include.h"

partial_post::partial_post (dlist *D, char *p, char *t, char *a, time_t d,
		    unsigned long c, cryptotype s, char z) : dlistNodeAuto (D)
{
	// StrDup-ready strings. not a kludge.
	point = p;
	title = t;
	author = a;
	date = d;
	crc = c;
	secret = s;
	zombie = z;
}

void partial_post::chpoint (char *n)
{
	char tmp [MAX_PRL];
	sprintf (tmp, "%s/%s", n, MsgIDofPRL (point));
	delete [] point;
	point = StrDup (tmp);
}

partial_post::~partial_post ()
{
	delete [] point; delete [] title; delete [] author;
}

////////////////////////////////////////////////////////////////////////////

post::post (unsigned int ID, char *t, char *par) : intNode (&Globals.Posts)
{
	Key = ID;
	rkey = (Key) ? cryptorand () : 0;
	_repsize = 0;
	title = StrDup (t);
	date = time (NULL);
	zombie = false;

	if (par) {
		parent = StrDup (par);
		if (selfref (par)) {
			secret = SECRET_WSELF;
			post *p = (post*)
				Globals.Posts.intFind (
					ahtoi (strrchr (par, '/') + 1));
			if (p) p->addreply (this);
		} else secret = SECRET_UNPOSTED;
	} else {
		char pself [MAX_PRL];
		sprintf (pself, "%s/%x.%x", Globals.ThisPoint, Key, rkey);
		parent = StrDup (pself);
		secret = SECRET_WSELF;
	}
}

post::post (unsigned int ID, FILE *F) : intNode (&Globals.Posts)
{
	load_ipi (F);
	fclose (F);
}

////////////////////////////////////////////////////////////////////////////

void post::addreply (char *p, char *t, char *a, time_t d, unsigned long c, cryptotype s)
{
	new partial_post (&replies, p, t, a, d, c, s);
	_repsize += strlen (p) + strlen (t) + strlen (a);
}

void post::addreply (post *p)
{
	char rp [MAX_PRL];

	sprintf (rp, "%s/%x.%x", Globals.ThisPoint, p->Key, p->rkey);
	new partial_post (&replies, StrDup (rp), p->title, Globals.ThisAuthor,
			  p->date, postcrc (p->Key), SECRET_WSELF);
	_repsize += strlen (p->title) + strlen (rp) + strlen (Globals.ThisAuthor);
	dump_ipi ();
}

void post::rmvreply (char *s)
{
	partial_post *p = (partial_post*) replies.Start;

	for (; p; p = (partial_post*) replies.Next (p))
		if (!strcmp (s, p->point)) {
			replies.dremove (p);
			_repsize -= strlen (p->point) + strlen (p->title) +
				    strlen (p->author);
			delete p;
			return;
		}
}

bool post::canpost (char *prl, char *point)
{
	if (replies.cnt >= MAXIMUM_REPLIES) return false;

	int c = 0, l = (point) ? strlen (point) : 0;

	partial_post *p = (partial_post*) replies.Start;
	for (; p; p = (partial_post*) replies.Next (p)) {
		if (!strcasecmp (prl, p->point)) return false;
		if ((point) && !strncasecmp (p->point, point, l)) ++c;
	}

	return c < MAXIMUM_REPLIES_PP;
}

bool post::pcrchange (cryptotype ok, char *np, cryptotype nk)
{
	partial_post *p = (partial_post*) replies.Start;

	for (; p; p = (partial_post*) replies.Next (p))
		if (p->secret == ok) {
			_repsize -= strlen (p->point);
			p->chpoint (np);
			_repsize += strlen (p->point);
			p->secret = nk;
			return true;
		}
	return false;
}

char *post::checksecret (cryptotype c)
{
	partial_post *p = (partial_post*) replies.Start;

	for (; p; p = (partial_post*) replies.Next (p))
		if (p->secret == c)
			return p->point;

	return NULL;
}

char *post::msgidsec (cryptotype c)
{
	partial_post *p = (partial_post*) replies.Start;

	for (; p; p = (partial_post*) replies.Next (p))
		if (p->secret == c)
			return (MsgIDofPRL (p->point));

	return NULL;	/* DANGER */
}

bool post::update_ready (cryptotype c)
{
	partial_post *p = (partial_post*) replies.Start;

	for (; p; p = (partial_post*) replies.Next (p))
		if (p->secret == c) {
			if (p->zombie) {
				p->zombie = false;
				return true;
			}
			return false;
		}

	return false;
}

bool post::needsupdate (cryptotype c)
{
	partial_post *p = (partial_post*) replies.Start;

	for (; p; p = (partial_post*) replies.Next (p))
		if (p->secret == c)
			return p->zombie;
	return false;
}

void post::updated (cryptotype o, cryptotype c)
{
	partial_post *p = (partial_post*) replies.Start;

	for (; p; p = (partial_post*) replies.Next (p))
		if (p->secret == o) {
			p->secret = c;
			break;
		}
}

bool post::pcpchange (cryptotype ok, char *np, cryptotype nk)
{
	if (secret != ok) return false;

	char foo [MAX_PRL];
	strcat (strcat (strcpy (foo, np), "/"), MsgIDofPRL (parent));

	delete [] parent;
	parent = StrDup (foo);
	secret = nk;
	return true;
}

unsigned int post::dbuffer_size ()
{
	return _repsize + 2 * MAX_MAX +
		replies.cnt * (4 + Globals.fieldOverhead + MAX_DATE + MAX_CRC);
}

void post::buildDOWNLOAD (o_argstream &O)
{
	char dateb [MAX_DATE];
	/* PROTOCOL-STUFF-LOCATION */
	O.setmax (dbuffer_size ());
	O.putarg (Globals.PMSGtext [PNET_OK]);
	O.putarg (title);
	O.putarg (parent);
	O.putarg (replies.cnt);

	partial_post *p = (partial_post*) replies.Start;

	for (; p; p = (partial_post*) replies.Next (p)) {
		O.putarg (p->point);
		O.putarg (Globals.PNetFields [PFIELD_TITLE],	p->title);
		O.putarg (Globals.PNetFields [PFIELD_AUTHOR],	p->author);
		O.putarg (Globals.PNetFields [PFIELD_DATE],	network_date (p->date, dateb));
		O.putarg (Globals.PNetFields [PFIELD_CRC],	p->crc);
		O.putzero ();
	}
}

/* ++----
 i/o with the filesystem -- .ipi format
----++ */

#define fwriteobj(o, f) fwrite (&o, sizeof o, 1, f)

void post::dump_ipi ()
{
	char tmpfile [30];
	sprintf (tmpfile, "%u.tmp", Key);
	FILE *F = fopen (tmpfile, "w");

	fwriteobj  (zombie,	F);
	fwriteobj  (rkey,	F);
	fwriteobj  (secret,	F);
	fputstring (parent,	F);
	fputstring (title,	F);
	fwriteobj  (date,	F);

	partial_post *p = (partial_post*) replies.Start;

	for (; p; p = (partial_post*) replies.Next (p)) {
		fputstring (p->point,	F);
		fputstring (p->title,	F);
		fputstring (p->author,	F);
		fwriteobj  (p->date,	F);
		fwriteobj  (p->crc,	F);
		fwriteobj  (p->secret,	F);
		fwriteobj  (p->zombie,	F);
	}

	fclose (F);

	rename_ipi (tmpfile, Key);
}

#define freadobj(o, f) fread (&o, sizeof o, 1, f)

void post::load_ipi (FILE *F)
{
	freadobj (zombie,	F);
	freadobj (rkey,		F);
	freadobj (secret,	F);
	parent	= fgetstring (MAX_PRL, F);
	title	= fgetstring (MAX_TITLE, F);
	freadobj (date,		F);

	char *p, *t, *a, st;
	cryptotype s;
	unsigned long c;
	time_t d;
	while ((p = fgetstring (MAX_PRL, F))) {
		t = fgetstring (MAX_TITLE, F);
		a = fgetstring (MAX_AUTHOR, F);
		freadobj (d,	F);
		freadobj (c,	F);
		freadobj (s,	F);
		freadobj (st,	F);
		new partial_post (&replies, p, t, a, d, c, s, st);
	}
}

/* ------------------------------
 * * * * * * * HACK * * * * * * * *
  	Editing Posts Manually
   ------------------------------ */

void post::Edit_Dpy ()
{
	char dateb [MAX_DATE];
	printf ("Post number "BB"%u"BP "  ("BB"%x.%x"BP")\n", Key, Key, rkey);
	printf ("secret="BB"%lx"BP" ("BB"%s"BP")\n", secret,
		(secret==SECRET_UNPOSTED)?"unposted"
		:((secret==SECRET_WSELF)?"reply to self":"posted"));
	printf ("zombie="BB"%i"BP"\n", zombie);
	printf ("title=["BB"%s"BP"]\n", title);
	printf ("date="BB"%lu    (%s)"BP"\n", date, network_date (date, dateb));
	printf ("In reply to ["BB"%s"BP"]\n", parent);
	printf (BB"%i"BP" replies:\n", replies.cnt);

	partial_post *p = (partial_post*) replies.Start;
	int i = 0;

	for (; p; p = (partial_post*) replies.Next (p)) {
		printf (" ["BB"%i"BP"]\tTitle=["BB"%s"BP"]\n", i++, p->title);
		printf ("\tSource=["BB"%s"BP"]\n", p->point);
		printf ("\tBy=["BB"%s"BP"]\n", p->author);
		printf ("\tAt=["BB"%lu (%s)"BP"]\n", p->date, network_date(p->date, dateb));
		printf ("\tCRC=["BB"%lx"BP"]\n", p->crc);
		printf ("\tSecret=["BB"%lx"BP"]\n", p->secret);
		printf ("\tzombie=["BB"%i"BP"]\n", p->zombie);
	}
}

void post::Edit ()
{
	char cmd [100];
	Edit_Dpy ();
	for (;;) {
		printf ("\n\n");
		if (!getuserline ("--- Command", cmd, 100)) break;
		if (cmd [0] == 'Q') break;
		if (!cmd [0]) {
			printf (BB BLI"d"BP BB"199"BP" to delete reply 199\n");
			printf (BB BLI"U"BP" mark as unposted\n");
			printf (BB BLI"D"BP" display\n");
			printf (BB BLI"S"BP" save changes and exit\n");
			printf (BB BLI"Q"BP" quit\n");
		} else if (cmd [0] == 'd') {
			unsigned int I = strtoul (cmd + 1, NULL, 10);
			partial_post *p = (partial_post*) replies.Start;

			for (; (I) && (p); p = (partial_post*) replies.Next (p))
				I--;
			if (I==0 && p!=NULL)
				replies.dremove (p);
		} else if (cmd [0] == 'U') {
			secret = SECRET_UNPOSTED;
			zombie = false;
		} else if (cmd [0] == 'D') {
			Edit_Dpy ();
		} else if (cmd [0] == 'S') {
			dump_ipi ();
			printf (" Saved.\n");
			break;
		}
	}
	puts ("\n");
}

// #################################################################
//	 			no members
//		(only no members can enter *this* club)
// #################################################################

void EditIPI (char *f)
{
	FILE *F;
	if ((F = fopen (f, "r"))) {
		Globals.Posts.intFind (strtoul (f, NULL, 10));
		post P (strtoul (f, NULL, 10), F);
		P.Edit ();
	} else printf ("No such file\n");
	exit (0);
}

bool newreply (post *P, char *p, char *m, char *t, char *a, unsigned int c, cryptotype s)
{
	char prl [MAX_PRL];

DPRINTF (DB_SLOT3, "!!!!!!!!!! NEW REPLY !!!!!!!\n");
	if (Banned (p)) return false;

	if (P->Key == 0 && !Globals.CanReplyToFinger) return false;

	sprintf (prl, "%s/%s", p, m);
	if (P->canpost (prl, p)) {
DPRINTF (DB_SLOT3, "!!!!!!!!!! ADDING TO REPLIES LIST !!!!!!\n");
		P->addreply (StrDup (prl), StrDup (t), StrDup (a),
			     time (NULL), c, s);
		RNewRep (P->Key, (partial_post*) P->replies.End);
		return true;
	}

	return false;
}

post *Submit (char *r, char *t, char *d, int l)
{
	post *p;
	char filename [10 + sizeof POST_TEXT_FILE], to [MAX_PRL];
	FILE *f;

DPRINTF (DB_SLOT3, "# # # # SUMBIT [%s] ===%i===\n", t, Globals.TopPostID);
	// crappy but efficient mechanism for "already posted" case
	if ((p = (post*) Globals.Posts.intFind (Globals.TopPostID - 1))
	&& !strcmp (p->title, t)) return NULL;

	sem_wait (&Globals.DatabaseLock);
	while (Globals.Posts.intFind (Globals.TopPostID))
		++Globals.TopPostID;

	p = new post (Globals.TopPostID, t, r);
	if (p->secret == SECRET_UNPOSTED)
		QNewReply (CPofPRL (r, to), p->Key);
	p->dump_ipi ();
	sem_post (&Globals.DatabaseLock);

	sprintf (filename, "%u" POST_TEXT_FILE, p->Key);

	if ((f = fopen (filename, "w"))) {
		fwrite (d, 1, l, f);
		fclose (f);
	}

	return p;
}

void Admin_Remove_Post (unsigned int ID, char *rprl)
{
	// this function only removes other people's posts.
	// to remove OUR posts is actually the same as
	// deleting the .ipi file. and this is safer
	post *p;

	if (!selfref (rprl)) {
		sem_wait (&Globals.DatabaseLock);
		if ((p = (post*) Globals.Posts.intFind (ID))) {
			p->rmvreply (rprl);
			RDelRep (ID, rprl);
		}
		sem_post (&Globals.DatabaseLock);
	}
}
